from nonebot import on_command
from nonebot.adapters.cqhttp import Bot
from nonebot.adapters.cqhttp.event import MessageEvent
from nonebot.typing import T_State

from ...database import DB
from ...utils import get_type_id, permission_check, to_me, handle_uid


dynamic_off = on_command('关闭动态', rule=to_me(), priority=5)
dynamic_off.__doc__ = """关闭动态 UID"""

dynamic_off.handle()(permission_check)

dynamic_off.handle()(handle_uid)

@dynamic_off.got('uid', prompt='请输入要关闭动态的UID')
async def _(bot: Bot, event: MessageEvent, state: T_State):
    """根据 UID 关闭动态"""

    async with DB() as db:
        if await db.set_sub('dynamic', False, uid=state['uid'],
                            type_=event.message_type,
                            type_id=get_type_id(event)):
            user = await db.get_user(state['uid'])
            assert user is not None
            await dynamic_off.finish(f"已关闭 {user.name}（{user.uid}）的动态推送")
        await dynamic_off.finish(f"UID（{state['uid']}）未关注，请先关注后再操作")


import json
from logging import exception
from hashlib import md5
from typing import Any, Dict
from urllib.parse import urlencode

import httpx
from httpx import ConnectTimeout, ReadTimeout, ConnectError
from nonebot.log import logger
from httpx._types import URLTypes


class RequestError(Exception):
    def __init__(self, code, message=None, data=None):
        self.code = code
        self.message = message
        self.data = data
    
    def __repr__(self):
        return f"<RequestError code={self.code} message={self.message}>"
    
    def __str__(self):
        return self.__repr__()


class BiliReq():
    def __init__(self):
        self.appkey = "4409e2ce8ffd12b8"
        self.appsec = "59b43e04ad6965f34319062b478f83dd"
        self.default_headers = {
            'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64)\
             AppleWebKit/537.36 (KHTML, like Gecko) Chrome/87.0.4280.88\
              Safari/537.36 Edg/87.0.664.60',
            'Referer': 'https://www.bilibili.com/'
        }
        # self.login = Config.get_login()
        self.proxies: Dict[URLTypes, Any] = {'all://': None}

    # TODO 制作一个装饰器捕获请求时的异常并用更友好的方式打印出来
    async def request(self, method, url, **kw) -> Dict:
        async with httpx.AsyncClient(proxies=self.proxies) as client:
            try:
                r = await client.request(method, url, **kw)
                r.encoding = 'utf-8'
                res: Dict = r.json()
            except ConnectTimeout:
                logger.error(f"连接超时（{url}）")
                raise
            except ReadTimeout:
                logger.error(f"接收超时（{url}）")
                raise
            except exception as e:
                logger.error(f"未知错误（url）")
                raise 
            
            if res['code'] != 0:
                raise RequestError(code=res['code'],
                                    message=res['message'],
                                    data=res.get('data'))
            return res['data']
    
    async def get(self, url, **kw):
        return await self.request('GET', url, **kw)

    async def post(self, url, **kw):
        return await self.request('POST', url, **kw)
    
    async def get_info(self, uid):
        url = f'https://api.bilibili.com/x/space/acc/info?mid={uid}'
        return await self.get(url, headers=self.default_headers)

    async def get_user_dynamics(self, uid):
        # need_top: {1: 带置顶, 0: 不带置顶}
        url = f'https://api.vc.bilibili.com/dynamic_svr/v1/dynamic_svr/space_history?host_uid={uid}&offset_dynamic_id=0&need_top=0'
        return await self.get(url, headers=self.default_headers)
    
    # async def get_new_dynamics(self):
    #     url = 'https://api.vc.bilibili.com/dynamic_svr/v1/dynamic_svr/dynamic_new'
    #     params = {'type_list': 268435455, 'access_key': self.login['access_token']}
        return await self.get(url, params=params)

    # async def get_history_dynamics(self, offset_id):
    #     url = 'https://api.vc.bilibili.com/dynamic_svr/v1/dynamic_svr/dynamic_history'
    #     params = {
    #         'type_list': 268435455,
    #         'access_key': self.login['access_token'],
    #         'offset_dynamic_id': offset_id
    #     }
    #     return await self.get(url, params=params)
    
    async def get_live_list(self, uids):
        """根据 UID 获取直播间信息列表"""

        url = 'https://api.live.bilibili.com/room/v1/Room/get_status_info_by_uids'
        data = {'uids': uids}
        return await self.post(url, headers=self.default_headers, data=json.dumps(data))
    
    # async def get_is_live_list(self):
    #     url = 'https://api.live.bilibili.com/xlive/app-interface/v1/relation/liveAnchor'
    #     params = {'access_key': self.login['access_token']}
    #     return await self.get(url, params=params)
    
    async def get_live_info(self, uid):
        url = f'https://api.live.bilibili.com/room/v1/Room/getRoomInfoOld?mid={uid}'
        return await self.get(url, headers=self.default_headers)

    def _sign(self, params):
        """获取 params 的 sign 值"""

        items = sorted(params.items())
        return md5(
            f"{urlencode(items)}{self.appsec}".encode('utf-8')
            ).hexdigest()



import base64
import shutil
from pathlib import Path
from typing import Optional

from appdirs import AppDirs
from nonebot import get_driver
from nonebot.log import logger
from playwright.async_api import Browser, async_playwright

from . import config

_browser: Optional[Browser] = None


async def init(**kwargs) -> Browser:
    global _browser
    browser = await async_playwright().start()
    _browser = await browser.chromium.launch(**kwargs)
    return _browser


async def get_browser(**kwargs) -> Browser:
    return _browser or await init(**kwargs)

 
async def get_dynamic_screenshot(url):
    browser = await get_browser()
    page = None
    try:
        page = await browser.new_page(device_scale_factor=2)
        await page.goto(url, wait_until='networkidle', timeout=10000)
        await page.set_viewport_size({"width": 2560, "height": 1080})
        card = await page.query_selector(".card")
        assert card
        clip = await card.bounding_box()
        assert clip
        bar = await page.query_selector(".text-bar")
        assert bar
        bar_bound = await bar.bounding_box()
        assert bar_bound
        clip['height'] = bar_bound['y'] - clip['y']
        image = await page.screenshot(clip=clip, full_page=True)
        await page.close()
        return base64.b64encode(image).decode()
    except Exception:
        if page:
            await page.close()
        raise


def install():
    """自动安装、更新 Chromium"""

    logger.info("正在检查 Chromium 更新")
    import sys
    from playwright.__main__ import main
    sys.argv = ['', 'install', 'chromium']
    try:
        main()
    except SystemExit:
        pass


def delete_pyppeteer():
    """删除 Pyppeteer 遗留的 Chromium"""

    dir = Path(AppDirs('pyppeteer').user_data_dir)
    if not dir.exists():
        return

    if not config.haruka_delete_pyppeteer:
        logger.info("检测到 Pyppeteer 依赖（约 300 M)，"
                    "新版 HarukaBot 已经不需要这些文件了。"
                    "如果没有其他程序依赖 Pyppeteer，请在 '.env.*' 中设置"
                    " 'HARUKA_DELETE_PYPPETEER=True' 并重启 Bot 后，将自动清除残留")
    else:
        shutil.rmtree(dir)
        logger.info("已清理 Pyppeteer 依赖残留")


get_driver().on_startup(delete_pyppeteer)
get_driver().on_startup(install)



import asyncio
import platform
from uvicorn.loops import asyncio as _asyncio
from uvicorn import config
from nonebot.log import logger


def asyncio_setup():
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)


@property
def should_reload(self):
    return False


if platform.system() == "Windows":
    _asyncio.asyncio_setup = asyncio_setup
    config.Config.should_reload = should_reload # type: ignore
    logger.warning("检测到当前为 Windows 系统，已自动注入猴子补丁")


import asyncio
import traceback
from datetime import datetime, timedelta

from nonebot.log import logger

from ...libs.bilireq import BiliReq
from ...database import DB
from ...libs.dynamic import Dynamic
from ...utils import safe_send, scheduler, get_dynamic_screenshot

last_time = {}

@scheduler.scheduled_job('interval', seconds=10, id='dynamic_sched')
async def dy_sched():
    """直播推送"""

    async with DB() as db:
        uid = await db.next_uid('dynamic')
        if not uid:
            return
        user = await db.get_user(uid)
        assert user is not None
        name = user.name

    logger.debug(f'爬取动态 {name}（{uid}）')
    br = BiliReq()
    dynamics = (await br.get_user_dynamics(uid)).get('cards', []) # 获取最近十二条动态
    # config['uid'][uid]['name'] = dynamics[0]['desc']['user_profile']['info']['uname']
    # await update_config(config)

    if len(dynamics) == 0: # 没有发过动态或者动态全删的直接结束
        return

    if uid not in last_time: # 没有爬取过这位主播就把最新一条动态时间为 last_time
        dynamic = Dynamic(dynamics[0])
        last_time[uid] = dynamic.time
        return
    
    for dynamic in dynamics[4::-1]: # 从旧到新取最近5条动态
        dynamic = Dynamic(dynamic)
        if dynamic.time > last_time[uid] and dynamic.time > datetime.now().timestamp() - timedelta(minutes=10).seconds:
            logger.info(f"检测到新动态（{dynamic.id}）：{name}（{uid}）")
            image = None
            for _ in range(3):
                try:
                    image = await get_dynamic_screenshot(dynamic.url)
                    break
                except Exception as e:
                    logger.error("截图失败，以下为错误日志:")
                    logger.error(traceback(e))
                await asyncio.sleep(0.1)
            if not image:
                logger.error("已达到重试上限，将在下个轮询中重新尝试")
            await dynamic.format(image)

            async with DB() as db:
                push_list = await db.get_push_list(uid, 'dynamic')
                for sets in push_list:
                    await safe_send(sets.bot_id, sets.type, sets.type_id, dynamic.message)

            last_time[uid] = dynamic.time
    await DB.update_user(uid, dynamic.name) # type: ignore



from nonebot import on_notice
from nonebot.adapters.cqhttp import Bot, GroupDecreaseNoticeEvent
from nonebot.typing import T_State

from ..database import DB


group_decrease = on_notice(priority=5)

@group_decrease.handle()
async def _(bot: Bot, event: GroupDecreaseNoticeEvent, state: T_State):
    """退群时，自动删除该群订阅列表"""

    if event.self_id == event.user_id:
        async with DB() as db:
            await db.delete_sub_list('group', event.group_id)



from typing import Union
from nonebot import on_command
from nonebot.adapters.cqhttp import Bot
from nonebot.adapters.cqhttp.event import (GroupMessageEvent, MessageEvent,
                                           PrivateMessageEvent)
from nonebot.permission import SUPERUSER
from nonebot.adapters.cqhttp.permission import GROUP_ADMIN, GROUP_OWNER
from nonebot.typing import T_State

from ...database import DB
from ...utils import to_me, handle_uid


at_on = on_command('开启全体', rule=to_me(), 
    permission=GROUP_OWNER | GROUP_ADMIN | SUPERUSER, 
    priority=5)
at_on.__doc__ = """开启全体 UID"""

at_on.handle()(handle_uid)

@at_on.got('uid', prompt='请输入要开启全体的UID')
async def _(bot: Bot, event: Union[PrivateMessageEvent, GroupMessageEvent],
            state: T_State):
    """根据 UID 开启全体"""

    if isinstance(event, PrivateMessageEvent):
        await at_on.finish("只有群里才能开启全体")
        return # IDE 快乐行
    async with DB() as db:
        if await db.set_sub('at', True, uid=state['uid'], type_='group',
                            type_id=event.group_id):
            user = await db.get_user(state['uid'])
            assert user is not None
            await at_on.finish(f"已开启 {user.name}（{user.uid}）直播推送的@全体")
        await at_on.finish(f"UID（{state['uid']}）未关注，请先关注后再操作")


from tortoise.models import Model
from tortoise.fields.data import CharField, IntField, BooleanField


class Sub(Model):
    type = CharField(max_length=10)
    type_id = IntField()
    uid = IntField()
    live = BooleanField()
    dynamic = BooleanField()
    at = BooleanField()
    bot_id = IntField()


class User(Model):
    uid = IntField(pk=True)
    name = CharField(max_length=20)


class Group(Model):
    id = IntField(pk=True)
    admin = BooleanField()


class Version(Model):
    version = CharField(max_length=30)



from nonebot import on_command
from nonebot.adapters.cqhttp import Bot
from nonebot.adapters.cqhttp.event import MessageEvent
from nonebot.typing import T_State

from ...database import DB
from ...utils import get_type_id, permission_check, to_me

sub_list = on_command('关注列表', aliases={'主播列表',}, rule=to_me(), priority=5)
sub_list.__doc__ = """关注列表"""

sub_list.handle()(permission_check)

@sub_list.handle()
async def _(bot: Bot, event: MessageEvent, state: T_State):
    """发送当前位置的订阅列表"""

    message = "关注列表（所有群/好友都是分开的）\n\n"
    async with DB() as db:
        subs = await db.get_sub_list(event.message_type, get_type_id(event))
        for sub in subs:
            user = await db.get_user(sub.uid)
            assert user is not None
            message += (
                f"{user.name}（{user.uid}）"
                f"直播：{'开' if sub.live else '关'}，"
                f"动态：{'开' if sub.dynamic else '关'}，"
                f"全体：{'开' if sub.at else '关'}\n"
            )
    await sub_list.finish(message)



from pathlib import Path
from typing import Union

import nonebot
from nonebot import require
from nonebot.adapters import Bot, Event
from nonebot.adapters.cqhttp import MessageEvent, MessageSegment
from nonebot.adapters.cqhttp.event import (GroupMessageEvent,
                                           PrivateMessageEvent)
from nonebot.adapters.cqhttp.exception import ActionFailed, NetworkError
from nonebot.adapters.cqhttp.permission import GROUP_ADMIN, GROUP_OWNER
from nonebot.exception import FinishedException
from nonebot.log import logger
from nonebot.permission import SUPERUSER
from nonebot.rule import Rule
from nonebot.typing import T_State

from .. import config


def get_path(*other):
    """获取数据文件绝对路径"""
    if config.haruka_dir:
        dir_path = Path(config.haruka_dir).resolve()
    else:
        dir_path = Path.cwd().joinpath('data')
        # dir_path = Path.cwd().joinpath('data', 'haruka_bot')
    return str(dir_path.joinpath(*other))


async def handle_uid(bot: Bot, event: MessageEvent, state: T_State):
    uid = event.get_plaintext().strip()
    if not uid:
        return
    if uid.isdecimal():
        state['uid'] = uid
    else:
        await bot.send(event, "UID 必须为纯数字")
        raise FinishedException


async def permission_check(bot: Bot,
                           event: Union[GroupMessageEvent, PrivateMessageEvent],
                           state: T_State):
    from ..database import DB
    if isinstance(event, PrivateMessageEvent):
        return
    async with DB() as db:
        if (await db.get_admin(event.group_id) and
            not await (GROUP_ADMIN | GROUP_OWNER | SUPERUSER)(bot, event)):
            await bot.send(event, "权限不足，目前只有管理员才能使用")
            raise FinishedException


def to_me():
    if config.haruka_to_me:
        from nonebot.rule import to_me
        return to_me()
    async def _to_me(bot: Bot, event: Event, state: T_State) -> bool:
        return True
    return Rule(_to_me)


async def safe_send(bot_id, send_type, type_id, message, at=False):
    """发送出现错误时, 尝试重新发送, 并捕获异常且不会中断运行"""
    
    try:
        bot = nonebot.get_bots()[str(bot_id)]
    except KeyError:
        logger.error(f"推送失败，Bot（{bot_id}）未连接")
        return
    
    if at and (await bot.get_group_at_all_remain(group_id=type_id)
               )['can_at_all']:
        message = MessageSegment.at('all') + message

    try:
        return await bot.call_api('send_'+send_type+'_msg', **{
        'message': message,
        'user_id' if send_type == 'private' else 'group_id': type_id
        })
    except ActionFailed as e:
        url = "https://haruka-bot.sk415.icu/usage/faq.html#机器人不发消息也没反应"
        logger.error(f"推送失败，账号可能被风控（{url}），错误信息：{e.info}")
    except NetworkError as e:
        logger.error(f"推送失败，请检查网络连接，错误信息：{e.msg}")

def get_type_id(event: MessageEvent):
    return (event.group_id if isinstance(event, GroupMessageEvent)
                           else event.user_id)



scheduler = require('nonebot_plugin_apscheduler')
assert scheduler is not None
scheduler = scheduler.scheduler


# bot 启动时检查 src\data\haruka_bot\ 目录是否存在
if not Path(get_path()).is_dir():
    Path(get_path()).mkdir(parents=True)

from . import patch
from .browser import get_dynamic_screenshot



from typing import Optional
from pydantic import BaseSettings


# 其他地方出现的类似 from .. import config，均是从 __init__.py 导入的 Config 实例
class Config(BaseSettings):
    
    haruka_dir: Optional[str] = None
    haruka_to_me: bool = True
    haruka_delete_pyppeteer: bool = False

    class Config:
        extra = "ignore"


from nonebot import on_command
from nonebot.adapters.cqhttp import Bot
from nonebot.adapters.cqhttp.event import MessageEvent
from nonebot.typing import T_State

from ...database import DB
from ...utils import get_type_id, permission_check, to_me, handle_uid


live_off = on_command('关闭直播', rule=to_me(), priority=5)
live_off.__doc__ = """关闭直播 UID"""

live_off.handle()(permission_check)

live_off.handle()(handle_uid)

@live_off.got('uid', prompt='请输入要关闭直播的UID')
async def _(bot: Bot, event: MessageEvent, state: T_State):
    """根据 UID 关闭直播"""

    async with DB() as db:
        if await db.set_sub('live', False, uid=state['uid'],
                            type_=event.message_type,
                            type_id=get_type_id(event)):
            user = await db.get_user(state['uid'])
            assert user is not None
            await live_off.finish(f"已关闭 {user.name}（{user.uid}）的直播推送")
        await live_off.finish(f"UID（{state['uid']}）未关注，请先关注后再操作")



from pathlib import Path

import click
import dotenv

env = {
    "HOST": "127.0.0.1",
    "PORT": "8080",
    "SECRET": '',
    "ACCESS_TOKEN": '',
    "SUPERUSERS": [],
    "COMMAND_START": [""],
    "HARUKA_DIR": "./data/"
}


def create_env():
    file_path = Path('./.env.prod').resolve()
    if file_path.exists():
        return

    while True:
        try:
            superusers = click.prompt('主人QQ号 (多个使用空格分开)', default='', show_default=False)
            superusers = [int(superuser) for superuser in superusers.split()]
            env['SUPERUSERS'] = superusers
            break
        except ValueError:
            print('输入格式有误，请重新输入')

    with open('.env.prod', 'w') as f:
        pass
    
    for key, value in env.items():
        dotenv.set_key(file_path, key, str(value).replace('\'', '\"').replace(' ', ''), quote_mode='never')

if __name__ == "__main__":
    create_env()



from nonebot.adapters.cqhttp.message import MessageSegment


class Dynamic():
    def __init__(self, dynamic):
        # self.origin = json.loads(self.card['origin'])
        self.dynamic = dynamic
        # self.card = json.loads(dynamic['card'])
        # self.dynamic['card'] = self.card
        self.type = dynamic['desc']['type']
        self.id = dynamic['desc']['dynamic_id']
        self.url = "https://t.bilibili.com/" + str(self.id)
        self.time = dynamic['desc']['timestamp']
        # self.origin_id = dynamic['desc']['orig_dy_id']
        self.uid = dynamic['desc']['user_profile']['info']['uid']
        self.name = dynamic['desc']['user_profile']['info'].get('uname')
        # self.name = dynamic['desc']['user_profile']['info'].get('uname', Config.get_name(self.uid))

    async def format(self, img):
        type_msg = {
            0: "发布了新动态",
            1: "转发了一条动态",
            8: "发布了新投稿",
            16: "发布了短视频",
            64: "发布了新专栏",
            256: "发布了新音频"
        }
        self.message = (f"{self.name} " +
                        f"{type_msg.get(self.type, type_msg[0])}：\n" +
                        f"{self.url}\n" +
                        MessageSegment.image(f"base64://{img}")
                        )



from nonebot.adapters.cqhttp.message import Message, MessageSegment
from nonebot.log import logger

from ...database import DB
from ...libs.bilireq import BiliReq
from ...utils import safe_send, scheduler

status = {}

@scheduler.scheduled_job('interval', seconds=10, id='live_sched')
async def live_sched():
    """直播推送"""

    async with DB() as db:
        uids = await db.get_uid_list('live')

    if not uids:
        return
    logger.debug(f'爬取直播列表，目前开播{sum(status.values())}人，总共{len(uids)}人')
    br = BiliReq()
    res = await br.get_live_list(uids)
    if not res:
        return
    for uid, info in res.items():
        new_status = 0 if info['live_status'] == 2 else info['live_status']
        if uid not in status:
            status[uid] = new_status
            continue
        old_status = status[uid]
        if new_status != old_status and new_status: # 判断是否推送过
            room_id = info['short_id'] if info['short_id'] else info['room_id']
            url = 'https://live.bilibili.com/' + str(room_id)
            name = info['uname']
            title = info['title']
            cover = (info['cover_from_user'] if info['cover_from_user']
                                             else info['keyframe'])
            logger.info(f"检测到开播：{name}（{uid}）")

            live_msg = (f"{name} 正在直播：\n{title}\n" + 
                        MessageSegment.image(cover) + f"\n{url}")
            async with DB() as db:
                push_list = await db.get_push_list(uid, 'live')
                for sets in push_list:
                    await safe_send(
                        bot_id = sets.bot_id,
                        send_type = sets.type,
                        type_id = sets.type_id,
                        message = live_msg,
                        at = sets.at
                    )
                await db.update_user(uid, name)
        status[uid] = new_status



from nonebot import get_driver
from nonebot.plugin.manager import PluginLoader

from .config import Config


if isinstance(globals()['__loader__'], PluginLoader):
    global_config = get_driver().config
    config = Config(**global_config.dict())
    from . import plugins

from .version import VERSION, __version__



from nonebot import on_request
from nonebot.adapters.cqhttp import Bot, FriendRequestEvent, GroupRequestEvent
from nonebot.typing import T_State


friend_req = on_request(priority=5)

@friend_req.handle()
async def friend_agree(bot: Bot, event: FriendRequestEvent, state: T_State):
    if str(event.user_id) in bot.config.superusers:
        await bot.set_friend_add_request(flag=event.flag, approve=True)


group_invite = on_request(priority=5)

@group_invite.handle()
async def group_agree(bot: Bot, event: GroupRequestEvent, state: T_State):
    if (event.sub_type == 'invite' and
        str(event.user_id) in bot.config.superusers):
        await bot.set_group_add_request(flag=event.flag, sub_type='invite',
                                        approve=True)


from nonebot import on_command
from nonebot.adapters.cqhttp import Bot
from nonebot.adapters.cqhttp.event import MessageEvent
from nonebot.typing import T_State

from ...database import DB
from ...utils import permission_check, to_me, get_type_id, handle_uid
from ...libs.bilireq import BiliReq, RequestError


add_sub = on_command('关注', aliases={'添加主播',}, rule=to_me(), priority=5)
add_sub.__doc__ = """关注 UID"""

add_sub.handle()(permission_check)

add_sub.handle()(handle_uid)

@add_sub.got('uid', prompt='请输入要关注的UID')
async def _(bot: Bot, event: MessageEvent, state: T_State):
    """根据 UID 订阅 UP 主"""

    uid = state['uid']
    async with DB() as db:
        user = await db.get_user(uid)
        name = user and user.name
    if not name:
        br = BiliReq()
        try:
            name = (await br.get_info(uid))['name']
        except RequestError as e:
            if e.code == -400 or e.code == -404:
                await add_sub.finish("UID不存在，注意UID不是房间号")
            elif e.code == -412:
                await add_sub.finish("操作过于频繁IP暂时被风控，请半小时后再尝试")
            else:
                await add_sub.finish(f"未知错误，请联系开发者反馈，错误内容：\n\
                                    {str(e)}")
    async with DB() as db:
        result = await db.add_sub(uid, event.message_type,
                                  get_type_id(event), event.self_id, name)
    if result:
        await add_sub.finish(f"已关注 {name}（{uid}）")
    await add_sub.finish(f"{name}（{uid}）已经关注了")


from typing import Union
from nonebot import on_command
from nonebot.adapters.cqhttp import Bot
from nonebot.adapters.cqhttp.event import GroupMessageEvent, PrivateMessageEvent
from nonebot.permission import SUPERUSER
from nonebot.adapters.cqhttp.permission import GROUP_ADMIN, GROUP_OWNER
from nonebot.typing import T_State

from ...database import DB
from ...utils import to_me


permission_on = on_command('开启权限', rule=to_me(), 
    permission=GROUP_OWNER | GROUP_ADMIN | SUPERUSER, 
    priority=5)
permission_on.__doc__ = """开启权限"""

@permission_on.handle()
async def _(bot: Bot, event: Union[PrivateMessageEvent, GroupMessageEvent],
            state: T_State):
    """开启当前群权限"""

    if isinstance(event, PrivateMessageEvent):
        await permission_on.finish("只有群里才能开启权限")
        return # IDE 快乐行
    async with DB() as db:
        if await db.set_permission(event.group_id, True):
            await permission_on.finish("已开启权限，只有管理员和主人可以操作")
        await permission_on.finish("权限已经开启了，只有管理员和主人可以操作")



from nonebot import on_command
from nonebot.adapters.cqhttp import Bot
from nonebot.adapters.cqhttp.event import MessageEvent
from nonebot.typing import T_State

from ...database import DB
from ...utils import get_type_id, permission_check, to_me, handle_uid


delete_sub = on_command('取关', aliases={'删除主播',}, rule=to_me(), priority=5)
delete_sub.__doc__ = """取关 UID"""

delete_sub.handle()(permission_check)

delete_sub.handle()(handle_uid)

@delete_sub.got('uid', prompt='请输入要取关的UID')
async def _(bot: Bot, event: MessageEvent, state: T_State):
    """根据 UID 删除 UP 主订阅"""

    uid = state['uid']
    async with DB() as db:
        name = getattr(await db.get_user(uid), 'name', None)
        if name:
            result = await db.delete_sub(uid, event.message_type,
                                         get_type_id(event))
        else:
            result = False

    if result:
        await delete_sub.finish(f"已取关 {name}（{uid}）")
    await delete_sub.finish(f"UID（{uid}）未关注")



from nonebot import on_command
from nonebot.adapters.cqhttp import Bot
from nonebot.adapters.cqhttp.event import MessageEvent
from nonebot.typing import T_State
from nonebot.matcher import matchers

from ..utils import to_me
from ..version import __version__


help = on_command('帮助', rule=to_me(), priority=5)

@help.handle()
async def test(bot: Bot, event: MessageEvent, state: T_State):
    message = "HarukaBot目前支持的功能：\n（请将UID替换为需要操作的B站UID）\n"
    for matchers_list in matchers.values():
        for matcher in matchers_list:
            if (matcher.plugin_name and
                matcher.plugin_name.startswith("haruka_bot") and
                matcher.__doc__):
                message += matcher.__doc__ + '\n'
    message += (f"\n当前版本：v{__version__}\n"
                "反馈&帮助群：629574472\n"
                "详细帮助：https://haruka-bot.sk415.icu/usage/")
    await help.finish(message)


import json
from pathlib import Path
from typing import Dict, List, Optional

from nonebot import get_driver
from nonebot.log import logger
from packaging.version import Version
from tortoise import Tortoise
from tortoise.query_utils import Q
from tortoise.queryset import QuerySet

from ..utils import get_path
from ..version import __version__
from .models import Group, Sub, User
from .models import Version as DBVersion

uid_list = {'live': {'list': [], 'index': 0},
            'dynamic': {'list': [], 'index': 0}}


class DB:
    """数据库交互类，与增删改查无关的部分不应该在这里面实现"""

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    @classmethod
    async def init(cls):
        from . import models
        await Tortoise.init(
            db_url=f"sqlite://{get_path('data.sqlite3')}",
            modules={'models': [locals()['models']]}
        )
        await Tortoise.generate_schemas()

    async def add_group(self, group_id, admin=True):
        """创建群设置"""
        
        if Group.filter(Q(id=group_id)).exists():
            # 权限开关已经创建过了
            return
        # TODO 自定义默认权限
        await Group.create(id=group_id, admin=admin)

    async def add_sub(self, uid, type_, type_id, bot_id, name, live=True,
                      dynamic=True, at=False) -> bool:
        """添加订阅"""
        
        if await self.get_sub(uid, type_, type_id):
            return False
        if not await self.get_user(uid):
            await self.add_user(uid, name)
        if type_ == 'group':
            await self.add_group(type_id)
        await Sub.create(
            uid=uid,
            type=type_,
            type_id=type_id,
            # TODO 自定义默认动态直播全体开关
            live=live,
            dynamic=dynamic,
            at=at,
            bot_id=bot_id
        )
        await self.update_uid_list()
        return True

    async def add_user(self, uid, name):
        """添加 UP 主信息"""

        await User.create(uid=uid, name=name)

    async def delete_group(self, group_id) -> bool:
        """删除群设置"""
        
        if await self.get_sub(type_='group', type_id=group_id):
            # 当前群还有订阅，不能删除
            return False
        await Group.filter(Q(id=group_id)).delete()
        return True

    async def delete_sub(self, uid, type_, type_id) -> bool:
        """删除指定订阅"""
        
        if not await self.get_sub(uid, type_, type_id):
            # 订阅不存在
            return False
        await self._get_subs(uid, type_, type_id).delete()
        await self.delete_user(uid)
        await self.update_uid_list()
        return True

    async def delete_sub_list(self, type_, type_id):
        "删除指定位置的推送列表"
        
        subs = self._get_subs(type_=type_, type_id=type_id)
        uids = [sub.uid async for sub in subs]
        await subs.delete()
        for uid in uids:
            await self.delete_user(uid)
        if type_ == 'group':
            await self.delete_group(type_id)
        await self.update_uid_list()

    async def delete_user(self, uid) -> bool:
        """删除 UP 主信息"""
        
        if await self.get_sub(uid):
            # 还存在该 UP 主订阅，不能删除
            return False
        await User.filter(Q(uid=uid)).delete()
        return True

    async def get_admin(self, group_id) -> bool:
        """获取指定群权限状态"""
        
        group = await Group.filter(Q(id=group_id)).first()
        if not group:
            return True
        return bool(group.admin)

    @classmethod
    async def get_name(cls, uid) -> Optional[str]:
        """获取 UP 主昵称"""

        user = await cls.get_user(uid)
        if user:
            return user.name
        return None

    async def get_push_list(self, uid, func) -> List[Sub]:
        """根据类型和 UID 获取需要推送的 QQ 列表"""
        
        return await self._get_subs(uid, **{func: True}).all()

    async def get_sub(self, uid=None, type_=None, type_id=None) -> Optional[Sub]:
        """获取指定位置的订阅信息"""
        
        return await self._get_subs(uid, type_, type_id).first()

    async def get_sub_list(self, type_, type_id) -> List[Sub]:
        """获取指定位置的推送列表"""
        
        return await self._get_subs(type_=type_, type_id=type_id).all()
        

    def _get_subs(self, uid=None, type_=None, type_id=None, live=None,
                       dynamic=None, at=None, bot_id=None) -> QuerySet[Sub]:
        """获取指定的订阅数据"""
        
        kw = locals()
        del kw['self']
        kw['type'] = kw.pop('type_')
        filters = [Q(**{key: value}) for key, value in kw.items()
                                     if value != None]
        return Sub.filter(Q(*filters, join_type='AND'))
        

    async def get_uid_list(self, func) -> List:
        """根据类型获取需要爬取的 UID 列表"""
        
        return uid_list[func]['list']        

    @classmethod
    async def get_user(cls, uid: int) -> Optional[User]:
        """获取 UP 主信息，没有就返回 None"""
        
        return await User.filter(Q(uid=uid)).first()

    async def _need_update(self):
        """根据版本号检查是否需要更新"""

        haruka_version = Version(__version__)
        db_version = await DBVersion.first()
        if not db_version:
            await DBVersion.create(version=__version__)
            return False
        db_version = Version(db_version.version)
        return haruka_version > db_version

    async def next_uid(self, func):
        """获取下一个要爬取的 UID"""
        
        func = uid_list[func]
        if func['list'] == []:
            return None

        if func['index'] >= len(func['list']):
            func['index'] = 1
            return func['list'][0]
        else:
            index = func['index']
            func['index'] += 1
            return func['list'][index]

    async def set_permission(self, group_id, switch) -> bool:
        """设置指定位置权限"""
        
        query = Group.filter(Q(id=group_id))
        group = await query.first()
        if not group:
            await Group.create(id=group_id, admin=switch)
            return True
        if group.admin == switch:
            return False
        await query.update(admin=switch)
        return True

    async def set_sub(self, conf, switch, uid=None, type_=None, type_id=None):
        """开关订阅设置"""

        subs = self._get_subs(uid, type_, type_id)
        if not await subs.exists():
            return False
        await subs.update(**{conf: switch})
        return True

    async def update_uid_list(self):
        """更新需要推送的 UP 主列表"""

        subs = Sub.all()
        uid_list['live']['list'] = list(set([sub.uid async for sub in subs
                                             if sub.live]))
        uid_list['dynamic']['list'] = list(set([sub.uid async for sub in subs
                                                if sub.dynamic]))

    @classmethod
    async def update_user(cls, uid: int, name: str) -> bool:
        """更新 UP 主信息"""

        return bool(await User.filter(Q(uid=uid)).update(name=name))

    async def update_version(self):
        """更新版本号"""

        if not await self._need_update():
            return
        
        logger.info("正在更新数据库")        
        DBVersion.all().update(version=__version__)
        logger.info(f"数据库已更新至 v{__version__}")

    async def migrate_from_json(self):
        """从 TinyDB 的 config.json 迁移数据"""

        if not Path(get_path('config.json')).exists():
            return
        
        logger.info("正在从 config.json 迁移数据库")
        with open(get_path('config.json'), 'r', encoding='utf-8') as f:
            old_db = json.loads(f.read())
        subs: Dict[int, Dict] = old_db['_default']
        groups: Dict[int, Dict] = old_db['groups']
        for sub in subs.values():
            await self.add_sub(
                uid = sub['uid'],
                type_ = sub['type'],
                type_id = sub['type_id'],
                bot_id = sub['bot_id'],
                name = sub['name'],
                live = sub['live'],
                dynamic = sub['dynamic'],
                at = sub['at']
            )
        for group in groups.values():
            await self.set_permission(group['group_id'], group['admin'])
        
        Path(get_path('config.json')).rename(get_path('config.json.bak'))
        logger.info("数据库迁移完成")

    async def backup(self):
        """备份数据库"""
        pass

    @classmethod
    async def get_login(cls):
        """获取登录信息"""
        pass

    @classmethod
    async def update_login(cls, tokens):
        """更新登录信息"""
        pass


async def init():
    async with DB() as db:
        await db.init()
        await db.update_version()
        await db.migrate_from_json()
        await db.update_uid_list()


get_driver().on_startup(init)
get_driver().on_shutdown(Tortoise.close_connections)
